package ru.dbotthepony.mc.otm.android

import net.minecraft.ChatFormatting
import net.minecraft.nbt.CompoundTag
import net.minecraft.network.chat.Component
import net.minecraft.world.item.ItemStack
import net.minecraftforge.common.util.INBTSerializable
import ru.dbotthepony.mc.otm.capability.MatteryPlayerCapability
import ru.dbotthepony.mc.otm.client.render.SkinElement
import ru.dbotthepony.mc.otm.core.readNbt
import ru.dbotthepony.mc.otm.container.set
import ru.dbotthepony.mc.otm.core.set
import ru.dbotthepony.mc.otm.core.writeNbt
import java.io.InputStream
import java.io.OutputStream

abstract class AndroidResearch(val type: AndroidResearchType<*>, val capability: MatteryPlayerCapability) : INBTSerializable<CompoundTag> {
	var isResearched = false
		protected set

	/**
	 * Whenever there are changes to network
	 */
	var isDirty = false

	/**
	 * Called when it is required to network everything again
	 */
	abstract fun invalidateNetwork()

	fun unResearch() {
		if (!isResearched) {
			return
		}

		onUnResearch()
		isResearched = false
		isDirty = true
	}

	abstract fun onUnResearch()
	abstract fun onResearched()

	/**
	 * Consumes required resources for researching this technology.
	 *
	 * Returns true if all required resources were consumed, false otherwise.
	 *
	 * This function MUST be atomic, hence it SHOULD NOT consume ANY resources if it returns false.
	 */
	abstract fun consumeResearchCost(simulate: Boolean): Boolean

	/**
	 * Grants all (or some) resources back consumed by [consumeResearchCost].
	 *
	 * Returns true whenever player accepted all resources refunded, false otherwise.
	 */
	abstract fun refund(simulate: Boolean): Boolean

	open fun writeNetwork(buff: OutputStream) {
		buff.writeNbt(serializeNBT())
	}

	open fun readNetwork(buff: InputStream) {
		deserializeNBT(buff.readNbt())
	}

	open val canResearch: Boolean get() {
		if (!consumeResearchCost(simulate = true)) {
			return false
		}

		if (!isPrerequisitesResearched || isAnyBlockerResearched) {
			return false
		}

		return true
	}

	/**
	 * Determines if this research is already blocked directly by another research.
	 */
	val isAnyBlockerResearched: Boolean get() {
		for (research in type.flatBlockedBy) {
			if (capability.getResearch(research).isResearched) {
				return true
			}
		}

		return false
	}

	/**
	 * Determines if this research is already blocked by another research, directly or indirectly.
	 *
	 * This is slower than [isAnyBlockerResearched] and makes sense only to check this for visualization purposes,
	 * since indirect blocking research will just turn [isPrerequisitesResearched] false (under normal conditions).
	 */
	val isAnyBlockerResearchedIndirect: Boolean get() {
		for (research in type.allBlockedBy) {
			if (capability.getResearch(research).isResearched) {
				return true
			}
		}

		return false
	}

	/**
	 * Determines if this research' direct prerequisites are researched.
	 */
	val isPrerequisitesResearched: Boolean get() {
		for (research in type.flatPrerequisites) {
			if (!capability.getResearch(research).isResearched) {
				return false
			}
		}

		return true
	}

	/**
	 * Determines if this research' prerequisites are researched, direct or indirect.
	 *
	 * Doesn't make much sense to check for this, since all prerequisites *should* have their prerequisites
	 * researched and so on.
	 *
	 * Strictly speaking, if someone tamper with research tree and mark all direct prerequisites unlocked
	 * without unlocking their prerequisites, [isPrerequisitesResearched] will return true, but *this* will return *false*.
	 */
	val isPrerequisitesResearchedIndirect: Boolean get() {
		for (research in type.allPrerequisites) {
			if (!capability.getResearch(research).isResearched) {
				return false
			}
		}

		return true
	}

	fun research(force: Boolean = false): Boolean {
		if (isResearched) {
			return false
		}

		if (force) {
			onResearched()
			isResearched = true
			isDirty = true
			return true
		} else if (consumeResearchCost(false)) {
			onResearched()
			isResearched = true
			isDirty = true
			return true
		}

		return false
	}

	/**
	 * List of all tooltip lines for this research
	 */
	open val tooltipLines: List<Component> by lazy {
		return@lazy listOf(type.displayName.copy().withStyle(ChatFormatting.WHITE))
	}

	/**
	 * List of all tooltip lines for this research, for local player at research screen (clientside only)
	 */
	open val screenTooltipLines: List<Component> by this::tooltipLines

	/**
	 * This should return whatever can be used by other code to display this research's name
	 */
	open val tooltipHeader: Component by lazy { tooltipLines[0] }

	/**
	 * This should return whatever can be used by other code to display this research's name,
	 * for local player at research screen (clientside only)
	 */
	open val screenTooltipHeader: Component by lazy { screenTooltipLines[0] }

	open val skinIcon: SkinElement? get() = null
	open val stackIcon: ItemStack? get() = null
	open val iconText: Component? get() = null

	override fun serializeNBT(): CompoundTag {
		return CompoundTag().also {
			it["researched"] = isResearched
		}
	}

	override fun deserializeNBT(nbt: CompoundTag) {
		isResearched = nbt.getBoolean("researched")
	}
}
