package ru.dbotthepony.mc.otm.network

import it.unimi.dsi.fastutil.bytes.ByteArrayList
import it.unimi.dsi.fastutil.io.FastByteArrayOutputStream
import net.minecraft.network.FriendlyByteBuf
import net.minecraft.network.chat.Component
import net.minecraft.world.item.ItemStack
import net.minecraftforge.network.NetworkDirection.PLAY_TO_CLIENT
import net.minecraftforge.network.NetworkDirection.PLAY_TO_SERVER
import net.minecraftforge.network.NetworkEvent
import org.apache.logging.log4j.LogManager
import ru.dbotthepony.mc.otm.android.AndroidFeature
import ru.dbotthepony.mc.otm.android.AndroidFeatureType
import ru.dbotthepony.mc.otm.android.AndroidResearch
import ru.dbotthepony.mc.otm.android.AndroidResearchType
import ru.dbotthepony.mc.otm.capability.MatteryPlayerCapability
import ru.dbotthepony.mc.otm.capability.matteryPlayer
import ru.dbotthepony.mc.otm.client.MatteryGUI
import ru.dbotthepony.mc.otm.client.minecraft
import ru.dbotthepony.mc.otm.core.ImpreciseFraction
import ru.dbotthepony.mc.otm.core.readImpreciseFraction
import ru.dbotthepony.mc.otm.core.writeImpreciseFraction
import ru.dbotthepony.mc.otm.menu.AndroidStationMenu
import ru.dbotthepony.mc.otm.registry.MRegistry
import java.io.ByteArrayInputStream
import java.io.OutputStream
import java.util.function.Supplier

class MatteryPlayerFieldPacket(val bytes: ByteArray, val length: Int) : MatteryPacket {
	constructor(stream: FastByteArrayOutputStream) : this(stream.array, stream.length)

	override fun write(buff: FriendlyByteBuf) {
		buff.writeBytes(bytes, 0, length)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true
		context.get().enqueueWork {
			val player = minecraft.player?.matteryPlayer ?: return@enqueueWork
			player.synchronizer.applyNetworkPayload(ByteArrayInputStream(bytes, 0, length))
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): MatteryPlayerFieldPacket {
			val readable = buff.readableBytes()
			return MatteryPlayerFieldPacket(ByteArray(readable).also(buff::readBytes), readable)
		}
	}
}

class AndroidResearchRequestPacket(val type: AndroidResearchType<*>) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeInt(MRegistry.ANDROID_RESEARCH.getID(type))
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true
		context.get().enqueueWork {
			val ply = context.get().sender ?: return@enqueueWork
			val android = ply.matteryPlayer ?: return@enqueueWork

			if (!android.isAndroid || ply.containerMenu !is AndroidStationMenu)
				return@enqueueWork

			android.getResearch(type).research()
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): AndroidResearchRequestPacket {
			return AndroidResearchRequestPacket(MRegistry.ANDROID_RESEARCH.getValue(buff.readInt()))
		}
	}
}

private fun getData(invoke: (stream: OutputStream) -> Unit): FastByteArrayOutputStream {
	val stream = FastByteArrayOutputStream()
	invoke(stream)
	return stream
}

class AndroidResearchSyncPacket(val type: AndroidResearchType<*>, val dataList: FastByteArrayOutputStream?, val dataBytes: ByteArray?) : MatteryPacket {
	constructor(feature: AndroidResearch) : this(feature.type, getData(feature::writeNetwork), null)

	override fun write(buff: FriendlyByteBuf) {
		dataList ?: throw NullPointerException("No byte list is present")
		buff.writeInt(MRegistry.ANDROID_RESEARCH.getID(type))
		buff.writeBytes(dataList.array, 0, dataList.length)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		dataBytes ?: throw NullPointerException("No data bytes array is present")

		context.get().packetHandled = true
		context.get().enqueueWork {
			val android = minecraft.player?.matteryPlayer ?: return@enqueueWork

			android.getResearch(type).readNetwork(ByteArrayInputStream(dataBytes))
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): AndroidResearchSyncPacket {
			return AndroidResearchSyncPacket(
				MRegistry.ANDROID_RESEARCH.getValue(buff.readInt()),
				null, ByteArray(buff.readableBytes()).also { buff.readBytes(it) }
			)
		}
	}
}

class AndroidFeatureSyncPacket(val type: AndroidFeatureType<*>, val dataList: FastByteArrayOutputStream?, val dataBytes: ByteArray?) : MatteryPacket {
	constructor(feature: AndroidFeature) : this(feature.type, getData(feature::writeNetwork), null)

	override fun write(buff: FriendlyByteBuf) {
		dataList ?: throw NullPointerException("No byte list is present")
		buff.writeInt(MRegistry.ANDROID_FEATURES.getID(type))
		buff.writeBytes(dataList.array, 0, dataList.length)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		dataBytes ?: throw NullPointerException("No data bytes array is present")

		context.get().packetHandled = true
		context.get().enqueueWork {
			val android = minecraft.player?.matteryPlayer ?: return@enqueueWork

			android.computeIfAbsent(type).readNetwork(ByteArrayInputStream(dataBytes))
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): AndroidFeatureSyncPacket {
			return AndroidFeatureSyncPacket(
				MRegistry.ANDROID_FEATURES.getValue(buff.readInt()),
				null, ByteArray(buff.readableBytes()).also { buff.readBytes(it) }
			)
		}
	}
}

class AndroidFeatureRemovePacket(val type: AndroidFeatureType<*>) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeInt(MRegistry.ANDROID_FEATURES.getID(type))
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true
		context.get().enqueueWork {
			val android = minecraft.player?.matteryPlayer ?: return@enqueueWork

			android.removeFeature(type)
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): AndroidFeatureRemovePacket {
			return AndroidFeatureRemovePacket(MRegistry.ANDROID_FEATURES.getValue(buff.readInt()))
		}
	}
}

class AndroidBatteryItemPacket(val item: ItemStack) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeItem(item)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true
		context.get().enqueueWork {
			val android = minecraft.player?.matteryPlayer ?: return@enqueueWork

			android.batteryItemStack = item
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): AndroidBatteryItemPacket {
			return AndroidBatteryItemPacket(buff.readItem())
		}
	}
}

class AndroidEnergyLevelPacket(val level: ImpreciseFraction, val isMaxEnergy: Boolean) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeImpreciseFraction(level)
		buff.writeBoolean(isMaxEnergy)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.get().packetHandled = true
		context.get().enqueueWork {
			val android = minecraft.player?.matteryPlayer ?: return@enqueueWork

			if (isMaxEnergy) {
				android.maxBatteryLevel = level
			} else {
				android.batteryLevel = level
			}
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): AndroidEnergyLevelPacket {
			return AndroidEnergyLevelPacket(buff.readImpreciseFraction(), buff.readBoolean())
		}
	}
}

class PlayerIterationPacket(val iteration: Int, val deathLog: List<Pair<Int, Component>>) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeInt(iteration)
		buff.writeInt(deathLog.size)

		for ((ticks, value) in deathLog) {
			buff.writeInt(ticks)
			buff.writeComponent(value)
		}
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true

		MatteryGUI.iteration = iteration
		MatteryGUI.deathLog.clear()
		MatteryGUI.deathLog.addAll(deathLog)
		MatteryGUI.showIterationUntil = System.currentTimeMillis() + 4000L
		MatteryGUI.showIterationUntilFade = System.currentTimeMillis() + 5000L
	}

	companion object {
		fun read(buff: FriendlyByteBuf): PlayerIterationPacket {
			val iteration = buff.readInt()
			val size = buff.readInt()

			val list = ArrayList<Pair<Int, Component>>()

			for (i in 0 until size) {
				list.add(buff.readInt() to buff.readComponent())
			}

			return PlayerIterationPacket(iteration, list)
		}
	}
}

class ExoSuitCarriedPacket(val itemStack: ItemStack, val containerState: Int) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeItem(itemStack)
		buff.writeInt(containerState)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true
		context.enqueueWork {
			val mattery = minecraft.player?.matteryPlayer ?: return@enqueueWork

			if (mattery.hasExoSuit) {
				mattery.exoSuitMenu.carried = itemStack
				mattery.exoSuitMenu.stateId = containerState
			}
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): ExoSuitCarriedPacket {
			return ExoSuitCarriedPacket(buff.readItem(), buff.readInt())
		}
	}
}

class ExoSuitSlotPacket(val slotId: Int, val itemStack: ItemStack, val containerState: Int) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeInt(slotId)
		buff.writeItem(itemStack)
		buff.writeInt(containerState)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true

		if (slotId < 0) {
			LOGGER.error("Unknown slot with ID {} in exosuit menu", slotId)
			return
		}

		context.enqueueWork {
			val mattery = minecraft.player?.matteryPlayer ?: return@enqueueWork

			if (mattery.hasExoSuit) {
				if (slotId >= mattery.exoSuitMenu.slots.size) {
					LOGGER.error("Unknown slot with ID {} in exosuit menu", slotId)
					return@enqueueWork
				}

				mattery.exoSuitMenu.slots[slotId].set(itemStack)
				mattery.exoSuitMenu.stateId = containerState
			}
		}
	}

	companion object {
		private val LOGGER = LogManager.getLogger()

		fun read(buff: FriendlyByteBuf): ExoSuitSlotPacket {
			return ExoSuitSlotPacket(buff.readInt(), buff.readItem(), buff.readInt())
		}
	}
}

class ExoSuitMenuInitPacket(val slots: List<ItemStack>, val carried: ItemStack, val containerState: Int) : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {
		buff.writeInt(slots.size)

		for (item in slots) {
			buff.writeItem(item)
		}

		buff.writeItem(carried)
		buff.writeInt(containerState)
	}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true
		context.enqueueWork {
			val mattery = minecraft.player?.matteryPlayer ?: return@enqueueWork

			if (mattery.hasExoSuit) {
				mattery.exoSuitMenu.initializeContents(containerState, slots, carried)
			}
		}
	}

	companion object {
		fun read(buff: FriendlyByteBuf): ExoSuitMenuInitPacket {
			val size = buff.readInt()
			val slots = ArrayList<ItemStack>(size)

			for (i in 0 until size) {
				slots.add(buff.readItem())
			}

			val carried = buff.readItem()
			val containerState = buff.readInt()

			return ExoSuitMenuInitPacket(slots, carried, containerState)
		}
	}
}

object ExoSuitMenuOpen : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true
		context.enqueueWork {
			val player = context.sender ?: return@enqueueWork
			val mattery = player.matteryPlayer ?: return@enqueueWork

			if (mattery.hasExoSuit) {
				player.containerMenu = mattery.exoSuitMenu
			}
		}
	}
}

object ExoSuitMenuClose : MatteryPacket {
	override fun write(buff: FriendlyByteBuf) {}

	override fun play(context: Supplier<NetworkEvent.Context>) {
		context.packetHandled = true
		context.enqueueWork {
			val player = context.sender ?: return@enqueueWork
			val mattery = player.matteryPlayer ?: return@enqueueWork

			if (player.containerMenu == mattery.exoSuitMenu) {
				player.containerMenu = player.inventoryMenu
			}
		}
	}
}

object MatteryPlayerNetworkChannel : MatteryNetworkChannel(
	version = "1",
	name = "player"
) {
	fun register() {
		add(MatteryPlayerFieldPacket::class, MatteryPlayerFieldPacket.Companion::read, PLAY_TO_CLIENT)

		add(AndroidResearchRequestPacket::class, AndroidResearchRequestPacket.Companion::read, PLAY_TO_SERVER)
		add(AndroidResearchSyncPacket::class, AndroidResearchSyncPacket.Companion::read, PLAY_TO_CLIENT)
		add(AndroidFeatureSyncPacket::class, AndroidFeatureSyncPacket.Companion::read, PLAY_TO_CLIENT)
		add(AndroidFeatureRemovePacket::class, AndroidFeatureRemovePacket.Companion::read, PLAY_TO_CLIENT)
		add(AndroidBatteryItemPacket::class, AndroidBatteryItemPacket.Companion::read, PLAY_TO_CLIENT)
		add(AndroidEnergyLevelPacket::class, AndroidEnergyLevelPacket.Companion::read, PLAY_TO_CLIENT)

		add(PlayerIterationPacket::class, PlayerIterationPacket.Companion::read, PLAY_TO_CLIENT)

		add(ExoSuitCarriedPacket::class, ExoSuitCarriedPacket.Companion::read, PLAY_TO_CLIENT)
		add(ExoSuitSlotPacket::class, ExoSuitSlotPacket.Companion::read, PLAY_TO_CLIENT)
		add(ExoSuitMenuInitPacket::class, ExoSuitMenuInitPacket.Companion::read, PLAY_TO_CLIENT)
		add(ExoSuitMenuOpen::class, { ExoSuitMenuOpen }, PLAY_TO_SERVER)
		add(ExoSuitMenuClose::class, { ExoSuitMenuClose }, PLAY_TO_SERVER)
	}
}
