package ru.dbotthepony.mc.otm.registry

import com.mojang.serialization.Lifecycle
import net.minecraft.core.BlockPos
import net.minecraft.core.Holder
import net.minecraft.core.Registry
import net.minecraft.core.WritableRegistry
import net.minecraft.resources.ResourceKey
import net.minecraft.resources.ResourceLocation
import net.minecraft.world.damagesource.DamageSource
import net.minecraft.world.entity.EntityType
import net.minecraft.world.level.BlockGetter
import net.minecraft.world.level.block.*
import net.minecraft.world.level.block.state.BlockBehaviour
import net.minecraft.world.level.block.state.BlockState
import net.minecraft.world.level.material.Material
import net.minecraft.world.level.material.MaterialColor
import net.minecraftforge.eventbus.api.IEventBus
import net.minecraftforge.fml.javafmlmod.FMLJavaModLoadingContext
import net.minecraftforge.registries.ForgeRegistry
import net.minecraftforge.registries.IForgeRegistry
import net.minecraftforge.registries.NewRegistryEvent
import net.minecraftforge.registries.RegistryBuilder
import ru.dbotthepony.mc.otm.OverdriveThatMatters
import ru.dbotthepony.mc.otm.android.AndroidFeatureType
import ru.dbotthepony.mc.otm.android.AndroidResearchType
import ru.dbotthepony.mc.otm.block.CargoCrateBlock
import ru.dbotthepony.mc.otm.registry.objects.ColoredDecorativeBlock
import ru.dbotthepony.mc.otm.registry.objects.CrateProperties
import ru.dbotthepony.mc.otm.registry.objects.DecorativeBlock
import ru.dbotthepony.mc.otm.registry.objects.StripedColoredDecorativeBlock
import java.util.function.Supplier
import kotlin.properties.ReadWriteProperty
import kotlin.reflect.KProperty

fun <T> Registry<T>.register(key: String, value: T): Holder<T> {
	return this.register(ResourceLocation(OverdriveThatMatters.MOD_ID, key), value)
}

fun <T> Registry<T>.register(key: ResourceLocation, value: T): Holder<T> {
	return (this as WritableRegistry<T>).register(ResourceKey.create(key(), key), value, Lifecycle.stable())
}

private class RegistryDelegate<T>(key: String) {
	private var value: Supplier<IForgeRegistry<T>?>? = null

	val location = ResourceLocation(OverdriveThatMatters.MOD_ID, key)
	val key: ResourceKey<Registry<T>> = ResourceKey.createRegistryKey(location)

	fun get(): IForgeRegistry<T> {
		val supp = value ?: throw IllegalStateException("Tried to access uninitialized registry $location")
		return supp.get() ?: throw IllegalStateException("Accessing registry $location too early")
	}

	fun build(event: NewRegistryEvent) {
		if (value != null) {
			throw IllegalStateException("Already built registry $location!")
		}

		value = RegistryBuilder<T>().let {
			it.setName(location)
			// it.type = AndroidFeatureType::class.java
			event.create(it)
		}
	}
}

class WriteOnce<V> : ReadWriteProperty<Any?, V> {
	private var value: V? = null

	override fun getValue(thisRef: Any?, property: KProperty<*>): V {
		return checkNotNull(value) { "Property ${property.name} is not initialized" }
	}

	override fun setValue(thisRef: Any?, property: KProperty<*>, value: V) {
		if (this.value != null) {
			throw IllegalStateException("Property ${property.name} already initialized")
		}

		this.value = value
	}
}

object MRegistry {
	private val features = RegistryDelegate<AndroidFeatureType<*>>("android_features")
	private val research = RegistryDelegate<AndroidResearchType<*>>("android_research")

	val ANDROID_FEATURES get() = features.get() as ForgeRegistry<AndroidFeatureType<*>>
	val ANDROID_RESEARCH get() = research.get() as ForgeRegistry<AndroidResearchType<*>>

	val ANDROID_FEATURES_LOCATION get() = features.location
	val ANDROID_RESEARCH_LOCATION get() = research.location

	val ANDROID_FEATURES_KEY get() = features.key
	val ANDROID_RESEARCH_KEY get() = research.key

	private fun register(event: NewRegistryEvent) {
		features.build(event)
		research.build(event)
	}

	val CARGO_CRATES = DecorativeBlock(MNames.CARGO_CRATE, ::CargoCrateBlock, baseItemGroup = OverdriveThatMatters.INSTANCE.CREATIVE_TAB)

	val TRITANIUM_BLOCK = DecorativeBlock.simple(MNames.TRITANIUM_BLOCK, {
		BlockBehaviour.Properties.of(Material.METAL, it?.materialColor ?: MaterialColor.COLOR_LIGHT_BLUE)
			.sound(SoundType.BASALT)
			.requiresCorrectToolForDrops()
			.explosionResistance(80f)
			.strength(4f)
	})

	val VENT = DecorativeBlock.simple(MNames.VENT, {
		BlockBehaviour.Properties.of(Material.METAL, it?.materialColor ?: MaterialColor.COLOR_LIGHT_BLUE)
			.sound(SoundType.BASALT)
			.requiresCorrectToolForDrops()
			.explosionResistance(20f)
			.strength(1.5f)
	})

	val VENT_ALTERNATIVE = DecorativeBlock.simple(MNames.VENT_ALTERNATIVE, {
		BlockBehaviour.Properties.of(Material.METAL, it?.materialColor ?: MaterialColor.COLOR_LIGHT_BLUE)
			.sound(SoundType.BASALT)
			.requiresCorrectToolForDrops()
			.explosionResistance(20f)
			.strength(1.5f)
	})

	val FLOOR_TILES = ColoredDecorativeBlock.simple(MNames.FLOOR_TILES, {
		BlockBehaviour.Properties.of(Material.STONE, it.materialColor)
			.sound(SoundType.STONE)
			.requiresCorrectToolForDrops()
			.strength(2f, 1.5f)
	})

	val UNREFINED_FLOOR_TILES = ColoredDecorativeBlock.simple(MNames.UNREFINED_FLOOR_TILES, {
		BlockBehaviour.Properties.of(Material.CLAY, it.materialColor)
			.sound(SoundType.GRAVEL)
			.strength(1f, 1f)
	})

	val INDUSTRIAL_GLASS = DecorativeBlock(MNames.INDUSTRIAL_GLASS, { color ->
		val properties = BlockBehaviour.Properties.of(Material.GLASS, if (color != null) color.materialColor else MaterialColor.NONE)
			.strength(1.5f, 35.0f)
			.requiresCorrectToolForDrops()
			.sound(SoundType.GLASS)
			.noOcclusion()
			.isValidSpawn { _: BlockState, _: BlockGetter, _: BlockPos, _: EntityType<*>? -> false }
			.isRedstoneConductor { _: BlockState, _: BlockGetter, _: BlockPos -> false }
			.isSuffocating { _: BlockState, _: BlockGetter, _: BlockPos -> false }
			.isViewBlocking { _: BlockState, _: BlockGetter, _: BlockPos -> false }

		if (color != null) {
			return@DecorativeBlock StainedGlassBlock(color, properties)
		}

		return@DecorativeBlock GlassBlock(properties)
	})

	val INDUSTRIAL_GLASS_PANE = DecorativeBlock(MNames.INDUSTRIAL_GLASS_PANE, { color ->
		val properties = BlockBehaviour.Properties.of(Material.GLASS, if (color != null) color.materialColor else MaterialColor.NONE)
			.strength(1.25f, 5.0f)
			.requiresCorrectToolForDrops()
			.sound(SoundType.GLASS)
			.noOcclusion()

		if (color != null) {
			return@DecorativeBlock StainedGlassPaneBlock(color, properties)
		}

		return@DecorativeBlock IronBarsBlock(properties)
	})

	val CRATE_RED = CrateProperties(MaterialColor.COLOR_RED, "crate_red")
	val CRATE_BLUE = CrateProperties(MaterialColor.COLOR_BLUE, "crate_blue")
	val CRATE_YELLOW = CrateProperties(MaterialColor.COLOR_YELLOW, "crate_yellow")
	val CRATE_GREEN = CrateProperties(MaterialColor.COLOR_GREEN, "crate_green")
	val CRATE_BLACK = CrateProperties(MaterialColor.COLOR_BLACK, "crate_black")
	val CRATE_PINK = CrateProperties(MaterialColor.COLOR_PINK, "crate_pink")
	val CRATE_PURPLE = CrateProperties(MaterialColor.COLOR_PURPLE, "crate_purple")

	val CRATE_LIST = listOf(
		CRATE_RED,
		CRATE_BLUE,
		CRATE_YELLOW,
		CRATE_GREEN,
		CRATE_BLACK,
		CRATE_PINK,
		CRATE_PURPLE,
	)

	const val DAMAGE_BECOME_ANDROID_ID = "otm_become_android"
	const val DAMAGE_BECOME_HUMANE_ID = "otm_become_humane"
	const val DAMAGE_EVENT_HORIZON_ID = "otm_event_horizon"
	const val DAMAGE_HAWKING_RADIATION_ID = "otm_hawking_radiation"
	const val DAMAGE_EXOSUIT_PROBE_ID = "otm_exosuit_probe"

	val DAMAGE_EXOSUIT_PROBE = ImmutableDamageSource(DamageSource(DAMAGE_EXOSUIT_PROBE_ID).bypassArmor().bypassMagic())

	val DAMAGE_BECOME_ANDROID = ImmutableDamageSource(DamageSource(DAMAGE_BECOME_ANDROID_ID).bypassArmor().bypassInvul().bypassMagic())
	val DAMAGE_BECOME_HUMANE = ImmutableDamageSource(DamageSource(DAMAGE_BECOME_HUMANE_ID).bypassArmor().bypassInvul().bypassMagic())
	val DAMAGE_EVENT_HORIZON = ImmutableDamageSource(DamageSource(DAMAGE_EVENT_HORIZON_ID).bypassMagic().bypassArmor())
	val DAMAGE_HAWKING_RADIATION = ImmutableDamageSource(DamageSource(DAMAGE_HAWKING_RADIATION_ID))
	const val DAMAGE_EMP_NAME = "otm_emp"
	const val DAMAGE_PLASMA_NAME = "otm_plasma"
	val DAMAGE_EMP: DamageSource = ImmutableDamageSource(EMPDamageSource())

	val TRITANIUM_STRIPED_BLOCK = StripedColoredDecorativeBlock(MNames.TRITANIUM_STRIPED_BLOCK, { colorA, _ ->
		Block(BlockBehaviour.Properties.of(Material.METAL, colorA.materialColor)
			.sound(SoundType.BASALT)
			.requiresCorrectToolForDrops()
			.explosionResistance(80f)
			.strength(4f))
	})

	fun initialize(bus: IEventBus) {
		bus.addListener(this::register)
		bus.addListener(MStats::registerVanilla)

		MBlocks.register(bus)
		MBlockEntities.register(bus)
		MEntityTypes.register(bus)
		MMenus.register(bus)
		MItems.register(bus)
		AndroidFeatures.register(bus)
		AndroidResearch.register(bus)
		MSoundEvents.register(bus)
		LootModifiers.register(bus)

		MRecipes.register(bus)
	}
}
